#include <tcl.h>
#ifdef LOGIC_TK
  #include <tk.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "copyrght.h"


int TkTest_Init(Tcl_Interp *interp);
int TclTest_Init(Tcl_Interp *interp);
int Tcl_AppInit(Tcl_Interp *interp);

/*(1) functions that define TCL commands are defined here */
/* int xxxx(ClientData, Tcl_Interp *, int, char **); */
int psl_subfilelist(ClientData, Tcl_Interp *, int, char **);
int psl_subfileload(ClientData, Tcl_Interp *, int, char **);
int psl_report(ClientData, Tcl_Interp *, int, char **);
int psl_element(ClientData, Tcl_Interp *, int, char **);
int psl_elcount(ClientData, Tcl_Interp *, int, char **);
int psl_next(ClientData, Tcl_Interp *, int, char **);
int psl_mapio(ClientData, Tcl_Interp *, int, char **);
int psl_getpdf1(ClientData, Tcl_Interp *, int, char **);
int psl_getpdf2(ClientData, Tcl_Interp *, int, char **);
int psl_findpeak(ClientData, Tcl_Interp *, int, char **);
int psl_storemap(ClientData, Tcl_Interp *, int, char **);
int psl_findnum(ClientData, Tcl_Interp *, int, char **);

/* Fortran subroutines used within the commands are defined in logiclib.h */
#include "logiclib.h"

int main (int argc, char *argv[])
{
#ifdef LOGIC_TK
  Tk_Main(argc, argv, Tcl_AppInit);
#else
  Tcl_Main(argc, argv, Tcl_AppInit);
#endif
  return 0;
}

/* initialize TCL */
int Tcl_AppInit(Tcl_Interp *interp) {
  /*  initialize LOGIC */
  if (PSL_INIT() != 0) return TCL_ERROR;
  /* Initialize TCL */
  if (Tcl_Init(interp) == TCL_ERROR) return TCL_ERROR;
  if (TclTest_Init(interp) == TCL_ERROR) return TCL_ERROR; 
#ifdef LOGIC_TK
  /* Initialize Tk */
  if (Tk_Init(interp) == TCL_ERROR) return TCL_ERROR;
  /* the syntax for the following changes as of Tcl 7.5 */
#if TCL_MAJOR_VERSION > 7 || (TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION >= 5)
  Tcl_SetVar(interp, "tcl_rcFileName", "~/.wishrc", TCL_GLOBAL_ONLY);
#else
  tcl_RcFileName = "~/.wishrc";
#endif
#endif
  return TCL_OK;
}  

/* (2) define the commands to be used */
int TclTest_Init(Tcl_Interp *interp) {
  Tcl_CreateCommand (interp, "sublist", psl_subfilelist,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "loadsub", psl_subfileload,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "report", psl_report,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "element", psl_element,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "elemcount", psl_elcount,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "nexthit", psl_next,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "savemap", psl_mapio,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "loadmap", psl_mapio,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "getpdf1", psl_getpdf1,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "getpdf2", psl_getpdf2,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "findpeak", psl_findpeak,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "storemap", psl_storemap,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "restoremap", psl_storemap,
		     (ClientData)NULL,NULL);
  Tcl_CreateCommand (interp, "findnumber", psl_findnum,
		     (ClientData)NULL,NULL);

  return Tcl_PkgProvide(interp, "logic", "1.0");
}

/*  interface for command sublist 

    This command lists the name associated with a subfile
    or all subfiles.

    Command syntax:
       sublist 
       sublist n
*/
int psl_subfilelist
    (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
    INTEGER4 i,j;
    int k;
    char string[40];

    if (argc == 1) 
      {
	j = 0;
	i = 1;
	j = PSL_TBLLIST(&i, string);
	while (j >= 0)
	  {
	    Tcl_AppendElement(interp, string);
	    i++;
	    j = PSL_TBLLIST(&i, string);
	  }
      }
    else if (argc == 2) 
      {
	/* i = atoi(argv[1]); */
	if (Tcl_GetInt(interp, argv[1], &k) != TCL_OK) 
	  {
	    Tcl_AppendResult(interp,
			     "\nusage:\n\t sublist\n(or)\t sublist <int>", 
			     (char *) NULL);
	    return TCL_ERROR;
	  }
	i = k; /* int to INTEGER4 */
	if (PSL_TBLLIST(&i, interp->result) < 0 ) 
	  {
	    Tcl_AppendResult(interp, "out of range subfile: ", argv[1], 
			     "\n", (char *) NULL);
	    return TCL_ERROR;
	  }
      }
    else
      {
	Tcl_AppendResult(interp,
			 "\nusage:\n\t sublist\n(or)\t sublist <int>", 
			 (char *) NULL);
	return TCL_ERROR;
      }
    return TCL_OK;
}

/*  interface for command loadsub

    This command lists the name associated with a subfile
    or all subfiles.

  Command syntax:
    loadsub -all
    loadsub -none
    loadsub -or  -subfile <int>
    loadsub -and -subfile <int>
    loadsub -remove -subfile <int>
    loadsub -not -subfile <int>
    loadsub -replace -subfile <int>
*/
int psl_subfileload
   (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  int error;			/* error flag */
  INTEGER4 ilogic,subfile,j;
  error = 0;
  j = 0;
  subfile = -1;
  ilogic = -1;
  if (argc == 2 && !strcmp(argv[1],"-all"))
    {
      ilogic = 6;
      subfile = 0;
    }
  else if (argc == 2 && !strcmp(argv[1],"-none"))
    {
      ilogic = 5;
      subfile = 0;
    }
  else if (argc == 4)
    while (j < 3)
    {
      j++;
      if (!strcmp(argv[j],"-or"))
	ilogic = 1;
      else if (!strcmp(argv[j],"-and"))
	ilogic = 0;
      else if (!strcmp(argv[j],"-remove"))
	ilogic = 2;
      else if (!strcmp(argv[j],"-not"))
	ilogic = 3;
      else if (!strcmp(argv[j],"-replace"))
	ilogic = 4;
      else if (!strcmp(argv[j],"-subfile"))
	{
	  j++;
	  subfile = atol(argv[j]);
	}
      else  /* unrecognized option */
	{
	  Tcl_AppendResult(interp,"unknown option: ",argv[j],"\n",
			   (char *) NULL);
	  error = 1;
	  j = 4;   /* terminate the loop */
	}
    }

  if (ilogic < 0 || subfile < 0) 
    {
	  Tcl_AppendResult(interp,"required option missing\n",
			   (char *) NULL);
      error = 1;
    }
  if (!error)
    {
    /*printf("Calling tblload with ilogic = %ld and subfile = %ld\n", */
    /* 	     ilogic, subfile); */
      j = PSL_TBLLOAD(&ilogic, &subfile);
      if (j < 0) 
	{
	  error = 1;
	  Tcl_AppendResult(interp,"error loading subfile\n",
			   (char *) NULL);
	}
      else
	sprintf(interp->result,"%ld",j );
    }
  if (error)
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp, "\t loadsub -all\n", (char *) NULL);
      Tcl_AppendResult(interp, "(or)\t loadsub -none\n", (char *) NULL);
      Tcl_AppendResult(interp,
	      "(or)\t loadsub [-or | -and | -remove | -not | -replace ]",
		       " -subfile <int>\n", (char *) NULL);
      return TCL_ERROR;
    }
  return TCL_OK;
}

/*  interface for command report

    This command reports information about the current table.

  Command syntax:
    report -total        returns the # of entries in the database
    report -current      returns the # of entries in the current table
    report -history      returns the history of the current table

*/
int psl_report
   (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  int i;
  INTEGER4 j;
  char string[101];
  if (argc == 2 && !strcmp(argv[1],"-total"))
    {
      i = 1;
      sprintf(interp->result,"%ld",PSL_REPHITS(&i) );
    }
  else if (argc == 2 && !strcmp(argv[1],"-current"))
    {
      i = 0;
      sprintf(interp->result,"%ld",PSL_REPHITS(&i) );
    }
  else if (argc == 2 && !strcmp(argv[1],"-history"))
    {
      i = 1;
      j = PSL_REPHISTORY(&i,string);
      while (j > 0)
	{
	  Tcl_AppendElement(interp, string);
	  i++;
	  j = PSL_REPHISTORY(&i,string);
	}
    }
  else
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp, "\t report -total\n", (char *) NULL);
      Tcl_AppendResult(interp, "(or)\t report -current\n", (char *) NULL);
      Tcl_AppendResult(interp, "(or)\t report -history", (char *) NULL);
      return TCL_ERROR;
    }
  return TCL_OK;
}


/*  interface for command element

    This command locates the entries that have a specified set of 
    elements. These entries are applied to the current table in
    a boolean operation.

  Command syntax:
    element [ -oneof e_list | -allof e_list ] 
            [ -alloptional | -optional e_list]
	    [ -or | -and | -remove | -not ]
    One option from each line must be present.

    Where -e_list is a list of elements by number {<int> <int>...}
    The list may be null, but the brackets or quotes must be present.
    Elements are defined as the integers from 1 to 103, and integers 
    from 104-121 represent groups of elements:
  1a	104              2a	105              3a	106           
  4a	107              5a	108              6a	109
  7a	110              8a	111              1b	112
  2b	113              3b	114              4b	115
  5b	116              6b	117              7b	118
  8b	119              LN	120              AD	121
*/
int psl_element
   (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  int error;			/* error flag */
  INTEGER4 i,j;
  INTEGER4 ilogic,reqval,optval;
  INTEGER4 reqelem[103],optelem[103];
  INTEGER4 reqgrp[18],optgrp[18];
  int k,listlen;
  char **listpt;
  error = 0;
  reqval = -1;
  optval = -1;
  ilogic = -1;
  for (j=0; j<18; j++) reqgrp[j] = optgrp[j] = 0;
  for (j=0; j<103; j++) reqelem[j] = optelem[j] = 0;

  j = 1;
  while (j < argc)
    {
      if (!strcmp(argv[j],"-allof") || !strcmp(argv[j],"-oneof"))
	{
	  if (reqval != -1) 
	    {
	      Tcl_AppendResult(interp,"option -allof or -oneof repeated\n",
			   (char *) NULL);
	      error = 1;
	    }
	  if (!strcmp(argv[j],"-allof"))
	    reqval = 0;
	  else
	    reqval = 1;
	  j++;
	  if (Tcl_SplitList(interp, argv[j], &listlen, &listpt) 
	      != TCL_OK) 
	    {
	      Tcl_AppendResult(interp,"error parsing list: ",argv[j],"\n",
			   (char *) NULL);
	      return TCL_ERROR;
	    }
	  for (i=0; i< listlen; i++)
	    {
	      if (Tcl_GetInt(interp, listpt[i], &k) != TCL_OK) 
		{
		  Tcl_AppendResult(interp,"error parsing number: ",listpt[i],
				   "\n",(char *) NULL);
		  error = 1;
		}
	      else if (k > 0 && k < 104) 
		reqelem[k-1] = 1;
	      else if (k > 103 && k < 122) 
		reqgrp[k-104] = 1;
	      else 
		{
		  Tcl_AppendResult(interp,"invalid value: ",listpt[i],"\n",
				   (char *) NULL);
		  error = 1;
		}
	    }
	  free ((char *) listpt);
	}
      else if (!strcmp(argv[j],"-optional") || !strcmp(argv[j],"-alloptional"))
	{
	  if (optval != -1) 
	    {
	      error = 1;
	      Tcl_AppendResult(interp,
			       "option -alloptional or -optional repeated\n",
			       (char *) NULL);
	    }
	  if (!strcmp(argv[j],"-alloptional"))
	    optval = 0;
	  else
	    {
	      optval = 1;
	      j++;
	      if (Tcl_SplitList(interp, argv[j], &listlen, &listpt) 
		  != TCL_OK) 
		{
		  Tcl_AppendResult(interp,"error parsing list: ",argv[j],"\n",
			   (char *) NULL);
		  return TCL_ERROR;
		}
	      for (i=0; i< listlen; i++)
		{
		  if (Tcl_GetInt(interp, listpt[i], &k) != TCL_OK) 
		    {
		      Tcl_AppendResult(interp,"error parsing number: ",
				       listpt[i],"\n",(char *) NULL);
		      error = 1;
		    }
		  else if (k > 0 && k < 104) 
		    optelem[k-1] = 1;
		  else if (k > 103 && k < 122) 
		    optgrp[k-104] = 1;
		  else 
		    {
		      Tcl_AppendResult(interp,"invalid value: ",listpt[i],"\n",
				       (char *) NULL);
		      error = 1;
		    }
		}
	      free ((char *) listpt);
	    }
	}
      else if (!strcmp(argv[j],"-or") || !strcmp(argv[j],"-and") ||
	       !strcmp(argv[j],"-remove") || !strcmp(argv[j],"-not") ||
	       !strcmp(argv[j],"-replace"))
	{
	  if (ilogic != -1) 
	    {
	      Tcl_AppendResult(interp,
		       "option -or -and -remove -not or -replace repeated\n",
			       (char *) NULL);
	      error = 1;
	    }
	  if (!strcmp(argv[j],"-and"))
	    ilogic = 0;
	  else if (!strcmp(argv[j],"-or"))
	    ilogic = 1;
	  else if (!strcmp(argv[j],"-remove"))
	    ilogic = 2;
	  else if (!strcmp(argv[j],"-not"))
	    ilogic = 3;
	  else if (!strcmp(argv[j],"-replace"))
	    ilogic = 4;
	}
      else
	{
	  Tcl_AppendResult(interp,"unknown option: ",argv[j],"\n",
			   (char *) NULL);
	  error = 1;
	  j = argc;
	}
      j++;
    }
  if (ilogic < 0 || reqval < 0 || optval <0 ) 
  {
    error = 1;
  }
  if (!error)
    {
/*       printf("elemsub ilogic = %ld, reqval = %ld, optval = %ld\n", */
/* 	     ilogic, reqval, optval); */
/*       printf("required elements: "); */
/*       for (j=0; j<103; j++) if (reqelem[j]) printf("%ld, ",j+1); */
/*       printf("\nrequired groups: "); */
/*       for (j=0; j<18; j++) if (reqgrp[j]) printf("%ld, ",j+1); */
/*       printf("\noptional elements: "); */
/*       for (j=0; j<103; j++) if (optelem[j]) printf("%ld, ",j+1); */
/*       printf("\noptional groups: "); */
/*       for (j=0; j<18; j++) if (optgrp[j]) printf("%ld, ",j+1); */
/*       printf("\n"); */

      j = PSL_ELEMSUB(&ilogic, &reqval, &optval,
			reqelem, reqgrp, optelem, optgrp);
      if (j < 0)
	{
	  Tcl_AppendResult(interp,"error processing request\n",
				   (char *) NULL);
	  error = 1;
	}
      else
	sprintf(interp->result,"%ld",j );
    }
  if (error)
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp, 
		       "element\t[ -oneof e_list | -allof e_list ]\n", 
		       (char *) NULL);
      Tcl_AppendResult(interp, 
		       "\t[ -alloptional | -optional e_list]\n",
		       (char *) NULL);
      Tcl_AppendResult(interp, 
		       "\t[ -or | -and | -remove | -not ]\n",
		       (char *) NULL);
      return TCL_ERROR;
    }
  return TCL_OK;
}
/*  interface for command nexthit

    This command is used to step through the current table.

  Command syntax:
    nexthit -first        sets direction + and returns the sequence 
                           of the first entry
    nexthit -last         sets direction - and returns the sequence
                           number of the last entry
    nexthit -forward      returns the sequence number of the next entry
                           in the +direction even after nexthit -last
    nexthit -backward     returns the sequence number of the previous entry
                           (in the +direction sense) even after nexthit -first
    nexthit               returns the sequence number of the next entry

  A negative value is returned if there are no more entries.
*/
int psl_next
   (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  static long direction = 0;
  INTEGER4 tmp_dir;
  static INTEGER4 iseq = 0;
  static INTEGER4 hitnum;
  INTEGER4 j;
  int k;
  int error;
  error = 0;
  tmp_dir = -1;
  if (argc == 2 && !strcmp(argv[1],"-first"))
    {
      direction = 0;
      iseq = 0;
    }
  else if (argc == 2 && !strcmp(argv[1],"-last"))
    {
      direction = 1;
      iseq = 0;
    }
  else if (argc == 2 && !strcmp(argv[1],"-forward"))
    tmp_dir = 0;
  else if (argc == 2 && !strcmp(argv[1],"-backward"))
    tmp_dir = 1;
  else if (argc == 2) 
    {
      /* i = atoi(argv[1]); */
      if (Tcl_GetInt(interp, argv[1], &k) != TCL_OK) 
	  {
	    Tcl_AppendResult(interp,
			     "invalid argument: argv[1]\n",
			     (char *) NULL);
	    error = 1;
	  }
      iseq = -1;
      hitnum = k;
    }
  else if (argc != 1)
    error = 1;
  
  if (error) 
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp, "\t nexthit -first\n", (char *) NULL);
      Tcl_AppendResult(interp, "(or)\t nexthit -last\n", (char *) NULL);
      Tcl_AppendResult(interp, "(or)\t nexthit -forward\n", (char *) NULL);
      Tcl_AppendResult(interp, "(or)\t nexthit -backward\n", (char *) NULL);
      Tcl_AppendResult(interp, "(or)\t nexthit <hitnum>\n", (char *) NULL);
      Tcl_AppendResult(interp, "(or)\t nexthit\n", (char *) NULL);
      return TCL_ERROR;
    }
  if (tmp_dir < 0) tmp_dir = direction;
  j = PSL_NEXTHIT(&iseq, &hitnum, &tmp_dir);
  sprintf(interp->result,"%ld",j);
  if (j > 0)
    iseq = j;
  else
    iseq = 0;
  return TCL_OK;
}

/*  interface for command savemap

    This command is used to save the current table as file <filename>.

  Command syntax:
    savemap <filename>


    interface for command loadmap

    This command is used to read a table saved on file <filename> 
    using savemap.

  Command syntax:
    loadmap <filename> [-or | -and | -remove | -not | -replace ]

*/
int psl_mapio
   (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  int error = 0;
  INTEGER4 ilogic = -1;
  int j,k;
  if (argc == 2 && !strcmp(argv[0],"savemap"))
      error = PSL_SAVEBITMAP(argv[1]);
  if (error) Tcl_AppendResult(interp,"error saving file\n",
			      (char *) NULL);
  else if (argc == 3 && !strcmp(argv[0],"loadmap"))
    {
      j = 1;
      k = 2;
      while (j < 3 && ilogic == -1)
	{
	  if (!strcmp(argv[j],"-or"))
	    ilogic = 1;
	  else if (!strcmp(argv[j],"-and"))
	    ilogic = 0;
	  else if (!strcmp(argv[j],"-remove"))
	    ilogic = 2;
	  else if (!strcmp(argv[j],"-not"))
	    ilogic = 3;
	  else if (!strcmp(argv[j],"-replace"))
	    ilogic = 4;
	  if (ilogic == -1 && j == 1)
	    {
	      j=2;
	      k=1;
	    }
	}
	  if (ilogic == -1)
	    {
	      Tcl_AppendResult(interp,"required option missing\n",
			   (char *) NULL);
	      error = 1;
	    }
	  else
	    {
	      error = PSL_READBITMAP(argv[k],&ilogic);
	      if (error) Tcl_AppendResult(interp,"error reading file\n",
					  (char *) NULL);
	    }
    }
  else
    error = 1;
  if (!error)
    return TCL_OK;

  if (!strcmp(argv[0],"savemap"))
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp, "\t savemap <filename>\n", (char *) NULL);
    }
  else
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp, "\tloadmap <filename> ",
		       "[-or | -and | -remove | -not | -replace ]\n",
		       (char *) NULL);
    }
  return TCL_ERROR;
}

/*  interface for command getpdf1

    This command returns the requested information for the entry # <int>.

  Command syntax:
    getpdf1 -seq <int> [-common] [-mineral] [-formula] [-chemname] 
                       [-reference] [-entry]
                       [-peaks [-wave <lambda> | -q ] -maxi ]

  The options listed below determine the information that is reported:
    -common        common name
    -mineral       mineral name
    -formula       chemical formula
    -chemname      chemical name
    -reference     literature reference
    -entry         JCPDS entry number in form ssnnnn where ss is set number
    -peaks         peaks as d-spaces and intensities 
    -wave <lambda> peaks as 2theta angles instead of d-spaces
    -q             peaks as Q values (2 pi/d-space) instead of d-spaces
    -maxi          peaks sorted by maximum intensity rather than d-space

    The command returns a list with the requested results in the order
    they are specified on the command line.
*/
int psl_getpdf1(ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  INTEGER4 iseq, ncrd, ndsp, maxi;
  INTEGER4 inten[255];
  REAL4 dsp[255],f;
  CHAR formula[256], minnam[256], commnam[256], chemnam[256];
  INTEGER4 year;
  CHAR coden[7], volume[6], page[6], author[256];
  REAL4 wave;

  int error,i,j,k;
  INTEGER4 l;
  double val;
  char string[2550]; /* need room for 255 peaks */
  char string1[50];

  error = 1;
  wave = 0.0;
  iseq = 0;
  maxi = 0;
  i=1;
  /* Pass one -- look for some qualifiers we will need 1st */
  while (i<argc)
    {
      if (i+1 < argc && !strcmp(argv[i],"-seq"))
	{
	  i++;
	  if (Tcl_GetInt(interp, argv[i], &k) == TCL_OK) iseq = k;
	}
      else if (i+1 < argc && !strcmp(argv[i],"-wave"))
	{
	  i++;
	  if (wave != 0) 
	    {
	      Tcl_AppendResult(interp,
			       "error: -wave or -q repeated or in conflict\n",
			       (char *) NULL);
	      error=1; /* flag error if there is a conflict */
	    }
	  if (Tcl_GetDouble(interp, argv[i], &val) == TCL_OK) wave = val;
	}
      else if (i < argc && !strcmp(argv[i],"-q")) 
	{
	  if (wave != 0.0) 
	    {
	      Tcl_AppendResult(interp,
			       "error: -wave or -q repeated or in conflict\n",
			       (char *) NULL);
	      error = 1; /* flag error if there is a conflict */
	    }
	  wave = -1.;
	}
      else if (i < argc && !strcmp(argv[i],"-maxi")) 
	maxi = 1;
      else if( strcmp(argv[i],"-peaks") && strcmp(argv[i],"-reference") &&
	       strcmp(argv[i],"-formula") && strcmp(argv[i],"-chemname") &&
	       strcmp(argv[i],"-mineral") && strcmp(argv[i],"-common") &&
	       strcmp(argv[i],"-entry"))
	{
	  Tcl_AppendResult(interp,"unknown option: ",argv[i],"\n",
			   (char *) NULL);
	  error = 1;
	}
      i++;
    }
  if (iseq > 0) error = 0;	/* found a sequence number -- continue */
  /* get the requested entry */
  if (!error)
    {
      l = PSL_REPPDF1(&iseq, &ncrd, &ndsp, inten, dsp, formula, minnam, 
			commnam, chemnam, &year, coden, volume, page, author);
      if (l < 0) 
	{
	  Tcl_AppendResult(interp,"error processing request\n",
				   (char *) NULL);
	  error = 1;
	}
    }
  i=1;
  /* get the rest of the qualifiers and process as we go */
  while (i<argc && !error)
    {
      if (!strcmp(argv[i],"-seq")) /* skip over */
	i++;
      else if (!strcmp(argv[i],"-wave")) /* skip over */
	i++;
      else if (!strcmp(argv[i],"-q")) /* skip over */
	{}
      else if (!strcmp(argv[i],"-maxi")) /* skip over */
	{}
      else if (!strcmp(argv[i],"-peaks")) /* list all peaks */
	{
	  strcpy(string, "");
	  /* sort by d-spaces */
	  if (maxi == 1) {
	    for (j=0; j<ndsp-1; j++) {
	      for (k=j+1; k<ndsp; k++) {
		if (inten[j] < inten[k]) {
		  f = dsp[j];
		  dsp[j] = dsp[k];
		  dsp[k] = f;
		  l = inten[j];
		  inten[j] = inten[k];
		  inten[k] = l;
		}
	      }
	    }
	  }
	  for (j=0; j<ndsp; j++)
	    {
	      if (wave == 0) {
		sprintf(string1, "%.3f %ld ",dsp[j], inten[j]);
		strcat(string,string1);
	      }
	      else if (wave < 0)
		{
		  REAL4 temp;
		  if (dsp[j] != 0.0)
		    {
		      temp = 2 * 3.14159 / dsp[j];
		      sprintf(string1, "%.3f %ld ", temp, inten[j]);
		      strcat(string,string1);
		    }
		}
	      else
		{
		  REAL4 temp=0.;
		  temp = PSL_PEAKFROMD(&dsp[j],&wave);
		  if (temp > 0.) {
		    sprintf(string1, "%.2f %ld ",temp, inten[j]);
		    strcat(string,string1);
		  }
		}
	    }
	    Tcl_AppendElement(interp, string);
	}
      else if (!strcmp(argv[i],"-reference")) 
	{
	  sprintf(string,"%s, %s, %s, p%s (%ld)",
		  author, coden, volume, page, year);
	  Tcl_AppendElement(interp, string);
	}
      else if (!strcmp(argv[i],"-formula"))
	Tcl_AppendElement(interp, formula);
      else if (!strcmp(argv[i],"-chemname"))
	Tcl_AppendElement(interp, chemnam);
      else if (!strcmp(argv[i],"-mineral"))
	Tcl_AppendElement(interp, minnam);
      else if (!strcmp(argv[i],"-common"))
	Tcl_AppendElement(interp, commnam);
      else if (!strcmp(argv[i],"-entry"))
	{
	  if (ncrd < 0) 
	    sprintf(string,"%ldD",-ncrd);
	  else
	    sprintf(string,"%ld",ncrd);
	  Tcl_AppendElement(interp, string);
	}
      else
	{
	  Tcl_AppendResult(interp,"unknown option: ",argv[i],"\n",
			   (char *) NULL);
	  error = 1;
	}
      i++;
    }
  if (!error)
    {}
  if (error)
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp, "\t getpdf1 -seq <int> ", (char *) NULL);
      Tcl_AppendResult(interp, "[-common] [-mineral] [-formula] [-chemname]\n",
		       (char *) NULL);
      Tcl_AppendResult(interp, "\t [-reference] [-entry]\n", (char *) NULL);
      Tcl_AppendResult(interp, "\t [ -peaks] [-wave <lambda> | -q ]", (char *) NULL);
      return TCL_ERROR;
    }
  return TCL_OK;
}
/*  interface for command getpdf2

    This command returns the AIDS*83 record for the entry # <int>.

  Command syntax:
    getpdf2 -seq <int>
or  getpdf2 -seq <int> -format <peak-unit> -file <filename>

  When -format is used the record is formatted and written on file <filename>
  values for <peak-unit> are D,Q,cr,fe,co,cu,mo,ag or a real number
  where D=d-space, Q=Q and the remaining values
  give 2theta values at the corresponding wavelength

*/

int psl_getpdf2(ClientData cData, Tcl_Interp *interp, int argc, char * argv[])
{
  int error;
  int k;
  INTEGER4 iseq;
  INTEGER4 iline;
  CHAR string[81];
  CHAR filename[81];
  CHAR format[21];

  error = 1;
  strcpy(format, "");
  strcpy(filename, "");
  if (argc == 3 && !strcmp(argv[1],"-seq"))
      if (Tcl_GetInt(interp, argv[2], &k) == TCL_OK) 
	{
	  iseq = k;
	  error = 0;
	}
  if (argc == 7)
    {
      int j;
      error = 3;
      j = 1;
      while (j < argc)
	{
	  /* get sequence number */
	  if (!strcmp(argv[j],"-seq"))
	    {
	      j++;
	      if (Tcl_GetInt(interp, argv[j], &k) == TCL_OK) 
		{
		  iseq = k;
		  error--;
		}
	    }
	  /* get format option (peak units) */
	  if (!strcmp(argv[j],"-format"))
	    {
	      j++;
	      strcpy(format, argv[j]);
	      error--;
	    }
	  /* get filename */
	  if (!strcmp(argv[j],"-file"))
	    {
	      j++;
	      strcpy(filename, argv[j]);
	      error--;
	    }
	  j++;
	}
    }

  if (error <= 0)
    {
      iline = 0;
      while(PSL_REPPDF2(&iseq, &iline, string, format, filename) > 0)
	{
	  if (iline > 0) Tcl_AppendElement(interp, string);
	  iline++;
	}
      /* if there are no lines then iseq must be invalid */
      if (iline == 0)
	{
	  sprintf(string,"%ld",iseq );
	  Tcl_AppendResult(interp, "Invalid sequence number: ", string,
			   (char *) NULL);
	  return TCL_ERROR;
	}
      return TCL_OK;
    }

  Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
  Tcl_AppendResult(interp, "\t getpdf2 -seq <int>\n", (char *) NULL);
  Tcl_AppendResult(interp, "  or\t getpdf2 -seq <int> -format <peak-unit> -file <filename>\n", (char *) NULL);
  return TCL_ERROR;
}

/*  interface for command elemcount

    This command locates the entries that have a specified range of
    unique elements. These entries are applied to the current table in
    a boolean operation.

  Command syntax:
        elemcount -count {<int> <int> ...} [ -or | -and | -remove | -not ]

*/
int psl_elcount
   (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  int error;			/* error flag */
  INTEGER4 j;
  INTEGER4 ilogic;
  INTEGER4 count_arr[10];
  int count;
  int i,k,listlen;
  char **listpt;

  error = 0;
  ilogic = count = -1;
  for (i=0; i< 10; i++) count_arr[i] = 0;

  j = 1;
  while (j < argc)
    {
      if (!strcmp(argv[j],"-count"))
	{
	  if (count != -1) 
	    {
	      Tcl_AppendResult(interp,"error: -count repeated\n",
			       (char *) NULL);
	      error = 1;
	    }
	  count = 1;
	  j++;
	  if (Tcl_SplitList(interp, argv[j], &listlen, &listpt) 
	      != TCL_OK) 
	    {
	      Tcl_AppendResult(interp,"error parsing list: ",
			       argv[j], "\n", (char *) NULL);
	      error = 1;
	    }
	  for (i=0; i< listlen; i++)
	    {
	      if (Tcl_GetInt(interp, listpt[i], &k) != TCL_OK) 
		{
		  Tcl_AppendResult(interp,"error parsing number: ",
				   listpt[i], "\n", (char *) NULL);
		  error = 1;
		}
	      else if (k >= 0 && k < 10) 
		count_arr[k] = 1;
	      else
		{
		  Tcl_AppendResult(interp,"invalid count value: ",listpt[i],
				   "\n",(char *) NULL);
		  error = 1;
		}
	    }
	}
      else if (!strcmp(argv[j],"-or"))
	ilogic = 1;
      else if (!strcmp(argv[j],"-and"))
	ilogic = 0;
      else if (!strcmp(argv[j],"-remove"))
	ilogic = 2;
      else if (!strcmp(argv[j],"-not"))
	ilogic = 3;
      else
	{
	  Tcl_AppendResult(interp,"unknown option: ",argv[j],"\n",
			   (char *) NULL);
	  error = 1;
	}
      j++;
    }
  if (count < 0 || ilogic < 0)
    {
      Tcl_AppendResult(interp,"required option missing\n",
		       (char *) NULL);
      error = 1;
    }
  if (!error)
    PSL_ELEMCOUNT(&ilogic, count_arr);
  if (error)
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp,
		       "\t elemcount -count {<int> <int> ...}",
		       " [ -or | -and | -remove | -not ]\n",
		       (char *) NULL);
      return TCL_ERROR;
    }
  return TCL_OK;
}

/*  interface for command findpeak

    This command locates the entries that have one or more peaks 
    a specified d-space range. The resulting entries are applied to 
    the current table in a boolean operation.

  Command syntax:
        findpeak -peak {<pos1> <pos2> ...} -error {<err1> <err2> ...}  
		 [-wave <lambda> | -q ]
	         [-any | -all] 
		 [-fast | -exact] 
		 [ -or | -and | -remove | -not ]
		 [-intmin <int> | -3strongest]
    where 
       -peak            precedes a list of peak positions as d-spaces (unless 
                        -wave or -q is used)
       -error           specifies the minimum error in the peak positions
       -wave <lambda>   positions are 2theta angles instead of d-spaces
       -q               positions are in Q (2 pi/d-space) instead of d-spaces
       -any             any one of the positions must be matched
       -all             all of the positions must be matched
       -intmin <int>    peaks must be above a threshold of <int>% 
       -3strongest      peaks must be one of the three strongest in the pattern
       -fast            expand the errors or lower "intmin" to speed the search
       -exact           use the limits exactly as specified
       -or -and -remove -not
                        logical options
*/
int psl_findpeak
   (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  int error;			/* error flag */
  int j, jpeak, jerr; 
  INTEGER4 allflag, intflag, fastflag, ilogic, numpeaks, retval;
  int k;
  REAL4 wave;
  REAL4 *peaklist = NULL;
  REAL4 *errlist = NULL;
  double val;

  error = 0;
  numpeaks = allflag = intflag = fastflag = ilogic = -1;
  wave = 0.0;
  jerr = jpeak = -1;
  j = 1;
  while (j < argc)
    {
      if (!strcmp(argv[j],"-peak"))
	{
	  int listlen;
	  char **listpt;
	  j++;
	  jpeak = j;
	  if (Tcl_SplitList(interp, argv[j], &listlen, &listpt) 
	      != TCL_OK) 
	    {
	      Tcl_AppendResult(interp,"error parsing list: ",argv[j],"\n",
			       (char *) NULL);
	      error = 1;
	    }
	  free ((char *) listpt);
	}
      else if (j+1 < argc && !strcmp(argv[j],"-wave"))
	{
	  j++;
	  if (wave != 0) 
	    {
	      Tcl_AppendResult(interp,
			       "error: -wave or -q repeated or in conflict\n",
			       (char *) NULL);
	      error=1; /* flag error if there is a conflict */
	    }
	  if (Tcl_GetDouble(interp, argv[j], &val) == TCL_OK) wave = val;
	}
      else if (!strcmp(argv[j],"-q")) 
	{
	  if (wave != 0.0) 
	    {
	      Tcl_AppendResult(interp,
			       "error: -wave or -q repeated or in conflict\n",
			       (char *) NULL);
	      error = 1; /* flag error if there is a conflict */
	    }
	  wave = -1.;
	}
      else if (j+1 < argc && !strcmp(argv[j],"-error"))
	jerr = ++j;  /* keep track of the error position */
      else if (!strcmp(argv[j],"-any") || !strcmp(argv[j],"-all"))
	{
	  if (allflag != -1) {
	    Tcl_AppendResult(interp,
			     "error: -any or -all repeated or in conflict\n",
			     (char *) NULL);
	    error=1; /* flag error if there is a conflict */
	  }
	   if (!strcmp(argv[j],"-any"))
	     allflag = 1;
	   else
	     allflag = 0;
	}
      else if (j+1 < argc && !strcmp(argv[j],"-intmin"))
	{
	  if (intflag != -1) 
	    {
	      Tcl_AppendResult(interp,
		     "error: -intmin or -3strongest repeated or in conflict\n",
			       (char *) NULL);
	      error=1; /* flag error if there is a conflict */
	    }	
	  j++;
	  if (Tcl_GetInt(interp, argv[j], &k) != TCL_OK) 
	    {
	      Tcl_AppendResult(interp,"error parsing number: ",argv[j],"\n",
			       (char *) NULL);
	      error = 1;
	    }
	  else
	    intflag = (k>100) ? 99 : k; /* limit the intensity to 99 */
	}
      else if (!strcmp(argv[j],"-3strongest"))
	{
	  if (intflag != -1) 
	    {
	      Tcl_AppendResult(interp,
		     "error: -intmin or -3strongest repeated or in conflict\n",
			       (char *) NULL);
	      error=1; /* flag error if there is a conflict */
	    }
	  intflag = 100;
	}
      else if (!strcmp(argv[j],"-fast") || !strcmp(argv[j],"-exact"))
	{
	  if (fastflag != -1) 
	    {
	      Tcl_AppendResult(interp,
		     "error: -fast or -exact repeated or in conflict\n",
			       (char *) NULL);
	      error=1; /* flag error if there is a conflict */
	    }
	  if (!strcmp(argv[j],"-fast")) 
	    fastflag = 0;
	  else
	    fastflag = 1;
	}
      else if (!strcmp(argv[j],"-or") || !strcmp(argv[j],"-and") ||
	       !strcmp(argv[j],"-remove") || !strcmp(argv[j],"-not"))
	{
	  if (ilogic != -1) 
	    {
	      Tcl_AppendResult(interp,
			       "option -or -and -remove or -not repeated\n",
			       (char *) NULL);
	      error = 1;
	    }
	  if (!strcmp(argv[j],"-and"))
	    ilogic = 0;
	  else if (!strcmp(argv[j],"-or"))
	    ilogic = 1;
	  else if (!strcmp(argv[j],"-remove"))
	    ilogic = 2;
	  else if (!strcmp(argv[j],"-not"))
	    ilogic = 3;
	}
      else
	{
	  Tcl_AppendResult(interp,"unknown option: ",argv[j],"\n",
			 (char *) NULL);
	  error = 1;
	}
      j++;
    }
  /* check that we have values for everything */
  if (allflag == -1 || intflag == -1 || fastflag == -1 || ilogic == -1 ||
      jpeak == -1 || jerr == -1) 
    {
      Tcl_AppendResult(interp,"required option missing\n",
		       (char *) NULL);
      error = 1;
    }
  /* now process the peaks and errors */
  if (!error) 
    {
      int listlen1, listlen2;
      char **listpt1, **listpt2;
      double val1, val2;
      if (Tcl_SplitList(interp, argv[jpeak], &listlen1, &listpt1) 
	  != TCL_OK) 
	{
	  Tcl_AppendResult(interp,"error parsing list: ",argv[jpeak],"\n",
			   (char *) NULL);
	  error = 1;
	}
      if (Tcl_SplitList(interp, argv[jerr], &listlen2, &listpt2) 
	  != TCL_OK) 
	{
	  Tcl_AppendResult(interp,"error parsing list: ",argv[jerr],"\n",
			   (char *) NULL);
	  error = 1;
	}
      /* allocate space for the intensity and error list */
      peaklist = (REAL4 *) malloc(listlen1 * sizeof(REAL4));
      errlist = (REAL4 *) malloc(listlen1 * sizeof(REAL4));
      numpeaks = 0;
      for (j=0; j< listlen1; j++)
	{
	  if (Tcl_GetDouble(interp, listpt1[j], &val1) != TCL_OK) 
	    {
	      Tcl_AppendResult(interp,"error parsing number: ",listpt1[j],"\n",
			   (char *) NULL);
	      error = 1;
	    }
	  if (j < listlen2) Tcl_GetDouble(interp, listpt2[j], &val2);
	  if (wave == 0)
	    {
	      peaklist[numpeaks] = val1;
	      errlist[numpeaks] = val2;
	      numpeaks++;
	    }
	  else 
	    {
	      INTEGER4 i;
	      REAL4 pos,pos1,pos2;
	      pos = val1;
	      i = 0;
	      peaklist[numpeaks] = PSL_PEAKTOD(&pos,&wave,&i);
	      if (peaklist[numpeaks] > 0)
		{
		  pos = val1 + val2;
		  pos1 = PSL_PEAKTOD(&pos,&wave,&i);
		  pos = val1 - val2;
		  pos2 = PSL_PEAKTOD(&pos,&wave,&i);
		  if (pos1 > 0 && pos2 > 0) 
		    errlist[numpeaks] = 0.5 * fabs (pos2 - pos1); 
		  else if (pos1 > 0)
		    errlist[numpeaks] = 0.5 * fabs (pos1 - peaklist[numpeaks]);
		  else
		    errlist[numpeaks] = 0.5 * fabs (pos2 - peaklist[numpeaks]);
		  numpeaks++;
		}
	      else
		Tcl_AppendResult(interp,"Ignoring out of range peak: ",
				 listpt1[j],"\n",(char *) NULL);
	    }
	}
      free ((char *) listpt1);
      free ((char *) listpt2);
    }
  if (!error)
    {
      retval = PSL_PEAKSUB(&numpeaks, peaklist, errlist, 
		  &allflag, &intflag, &fastflag, &ilogic, &wave); 
/*       for (j=0;j<numpeaks; j++) */
/* 	printf("%f %f ",peaklist[j], errlist[j]); */
/*       printf( */
/*   "\nnumpeaks=%ld, allflag=%ld, intflag=%ld, fastflag=%ld, ilogic=%ld\n", */
/*        numpeaks, allflag, intflag, fastflag, ilogic); */
    }
  else
    retval = 0;
  free ((char *) peaklist);
  free ((char *) errlist);
  if (retval < 0) 
    {
      Tcl_AppendResult(interp, "Error in peak input\n", (char *) NULL);
      return TCL_ERROR;
    }
  if (error)
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp,
        "\t findpeak -peak {<pos1> <pos2> ...} -error {<err1> <err2> ...}\n",
	"\t\t [-wave <lambda> | -q ]\n",
	"\t\t [-any | -all]\n", 
	"\t\t [-fast | -exact]\n", 
	"\t\t [-intmin <int> | -3strongest]\n",
	"\t\t [ -or | -and | -remove | -not ]\n",
		       (char *) NULL);
      return TCL_ERROR;
    }
  return TCL_OK;
}

/*  interface for command storemap

    This command stores the current table for potential reuse 
    in the undo command or restores the saved information.

    Command syntax:
       storemap
       restoremap
*/

int psl_storemap
    (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
    if (argc == 1 && !strcmp(argv[0],"storemap") ) 
      PSL_STOREBITMAP(0L);	/* the argument is not used at this time */
    else if (argc == 1 && !strcmp(argv[0],"restoremap") )
      PSL_RESTOREBITMAP(0L);	/* the argument is not used at this time */
    else
      {
	Tcl_AppendResult(interp,
			 "\nusage:\n\tstoremap\n(or)\trestoremap ", 
			 (char *) NULL);
	return TCL_ERROR;
      }
    return TCL_OK;
}

/*  interface for command findnumber

    This command locates the entries that have PDF numbers within 
    a specified range or in a specified list.
    The resulting entries are applied to the current table in a 
    boolean operation.

  Command syntax:
        findnumber
	         [-list {<pos1> <pos2> ...}]
		 [-range <start> <end>]
		 [ -or | -and | -remove | -not ]
    where 
       -list            precedes a list of entry numbers
       -range           precedes a list a starting and ending entry number
       -or -and -remove -not
                        logical options
*/

int psl_findnum
   (ClientData cData, Tcl_Interp *interp, int argc, char * argv[]) 
{
  INTEGER4 ilogic, retval, num1, num2;
  int j, list, range, error;
  j = 1;
  error = 0;
  range = list = ilogic = -1;
  while (j < argc)
    {
      if (!strcmp(argv[j],"-list")) 
	{
	  int listlen,i,k;
	  char **listpt;
	  j++;
	  list = j;
	  if (Tcl_SplitList(interp, argv[j], &listlen, &listpt) 
	      != TCL_OK) 
	    {
	      Tcl_AppendResult(interp,"error parsing list: ",argv[j],"\n",
			       (char *) NULL);
	      error = 1;
	    }
	  for (i=0; i< listlen; i++)
	    {
	      if (Tcl_GetInt(interp, listpt[i], &k) != TCL_OK) 
		{
		  Tcl_AppendResult(interp,"error parsing number: ",listpt[i],
				   "\n",(char *) NULL);
		  error = 1;
		}
	      else if (k < 0 || k > 999999) 
		{
		  Tcl_AppendResult(interp,"invalid entry number: ",listpt[i],
				   "\n",(char *) NULL);
		  error = 1;
		}
	    }
	  free ((char *) listpt);
	}
      else if (!strcmp(argv[j],"-range")) 
	{
	  range = 1;
	  if (j+2 >= argc) {
	    Tcl_AppendResult(interp,"missing numbers after -range\n",(char *) NULL);
	    error = 1;
	  }
	  else {
	    int k;
	    j++;
	    if (Tcl_GetInt(interp, argv[j], &k) != TCL_OK) {
	      Tcl_AppendResult(interp,"error parsing number: ",argv[j],
			       "\n",(char *) NULL);
	      error = 1;
	    }
	    num1 = k;
	    j++;
	    if (Tcl_GetInt(interp, argv[j], &k) != TCL_OK) {
	      Tcl_AppendResult(interp,"error parsing number: ",argv[j],
			       "\n",(char *) NULL);
	      error = 1;
	    }
	    num2 = k;
	  }
	}
      else if (!strcmp(argv[j],"-or") || !strcmp(argv[j],"-and") ||
	       !strcmp(argv[j],"-remove") || !strcmp(argv[j],"-not"))
	{
	  if (ilogic != -1) 
	    {
	      Tcl_AppendResult(interp,
			       "option -or -and -remove or -not repeated\n",
			       (char *) NULL);
	      error = 1;
	    }
	  if (!strcmp(argv[j],"-and"))
	    ilogic = 0;
	  else if (!strcmp(argv[j],"-or"))
	    ilogic = 1;
	  else if (!strcmp(argv[j],"-remove"))
	    ilogic = 2;
	  else if (!strcmp(argv[j],"-not"))
	    ilogic = 3;
	}
      else
	{
	  Tcl_AppendResult(interp,"unknown option: ",argv[j],"\n",
			 (char *) NULL);
	  error = 1;
	}
      j++;
    }
  /* check that we have values for everything */
  if ((list == -1 && range == -1) || ilogic == -1 )
    {
      Tcl_AppendResult(interp,"required option missing\n",
		       (char *) NULL);
      error = 1;
    }
  else if (list != -1 && range != -1)
    {
      Tcl_AppendResult(interp,"-list and -range cannot be used together\n",
		       (char *) NULL);
      error = 1;
    }
  /* now process the list */
  retval = -1;
  if ( list != -1  && !error) {
    int listlen,i,k;
    char **listpt;
    INTEGER4 *numlist;
    INTEGER4 numentries;
    if (Tcl_SplitList(interp, argv[list], &listlen, &listpt) != TCL_OK) {
      Tcl_AppendResult(interp,"error parsing list: ",argv[list],"\n", (char *) NULL);
      error = 1;
    }
    numlist = (INTEGER4 *) malloc(listlen * sizeof(INTEGER4));
    numentries = listlen;
    for (i=0; i< listlen; i++) {
      if (Tcl_GetInt(interp, listpt[i], &k) != TCL_OK) {
	Tcl_AppendResult(interp,"error parsing number: ",listpt[i],
			 "\n",(char *) NULL);
	error = 1;
      }
      else if (k >= 0 && k <= 999999) 
	numlist[i] = k;
      else {
	Tcl_AppendResult(interp,"invalid value: ",listpt[i],"\n",
			 (char *) NULL);
	error = 1;
      }
    }
    if (!error) retval = PSL_SETBYNUMBER(&ilogic, numlist, &numentries);
    free ((char *) numlist);
    free ((char *) listpt);
  }

  if ( range != -1 && !error) 
    retval = PSL_SETBYRANGE(&ilogic, &num1, &num2);
  
  if (retval == 0) {
    Tcl_AppendResult(interp, "No valid entries were found\n", (char *) NULL);
    return TCL_ERROR;
  }

  if (error)
    {
      Tcl_AppendResult(interp, "\nusage:\n", (char *) NULL);
      Tcl_AppendResult(interp,
        "\t findnumber\n\t\t-list {<ent#1> <ent#2> ...}\n (or)\t\t-range <start> <end>\n",
	"\t\t [ -or | -and | -remove | -not ]\n",
		       (char *) NULL);
      return TCL_ERROR;
    }
  sprintf(interp->result,"%ld",retval);
  return TCL_OK;
}

/* (3) add a subroutine */
/* (4) remember to add a entry in LOGIC.doc */
